# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Barcelona Supercomputing Center

# @license: https://www.gnu.org/licenses/gpl-3.0.html
# @author: see AUTHORS file

from netCDF4 import Dataset as nc
from netCDF4 import MFDataset as mnc
from grads import GaLab
import pandas as pd
import numpy as np
from datetime import datetime
from datetime import timedelta
import os
import os.path


class MapData(object):
    """Wrapper class to handle all the data needed to draw a map"""

    def __init__(self, **kwargs):
        print "mapData initializer"
        self.reset()
        self.maxReset()
        # Process optional arguments
        if(kwargs.has_key('map_data')):
            self.map_data = kwargs['map_data'];

    def reset(self):
        self.map_data = None
        self.wind_data = None
        self.contour_data = None
        self.scatter_data = None

    def maxReset(self):
        self.max_data = None

    def setGridData(self, data):
        #print "Setting grid data"
        self.grid_data = data

    def setMapData(self, data):
        #print "Setting map data"
        self.map_data = data

    def setMaxData(self, data):
        self.max_data = data

    def setWindData(self, data):
        self.wind_data = data

    def setContourData(self, data):
        self.contour_data = data

    def setScatterData(self, data):
        self.scatter_data = data

    def setMaxData(self, data):
        self.max_data = data

    def getGridData(self):
        return self.grid_data

    def getMapData(self):
        return self.map_data

    def getWindData(self):
        return self.wind_data

    def getContourData(self):
        return self.contour_data

    def getMaxData(self):
        return self.max_data

    def getScatterData(self):
        return self.scatter_data

    def hasMapData(self):
        return (self.map_data != None)

    def hasWindData(self):
        return (self.wind_data != None)

    def hasContourData(self):
        return self.contour_data is not None

    def hasMaxData(self):
        return (self.max_data != None)

    def hasScatterData(self):
        return (self.scatter_data != None)


class MapDataHandler(object):
#    ncdf = None
#    data = None
#    srcs = []
#    srcgaps = []
#    windsrc = []
#    windopts = []
#    indir = None
#    lats = []
#    lons = []
#    max = None
#    transf = False

    def __init__(self, sources, srcvars, indir, srcgaps, windsource, windopts, lats, lons, transf, **kwargs):
        self.galab = GaLab(Bin='grads', Window=False, Echo=False)
        self.ncdf = [[],[]] # list of lists, srcfiles and windfiles
        self.srcs = sources
        self.srcvars = srcvars
        self.srcgaps = srcgaps
        self.indir = indir
        self.windsrc = windsource
        self.windopts = windopts
        self.lats = lats
        self.lons = lons
        self.maxdata = None
        self.lastTstep = None
        self.transf = transf
        self.data = MapData()
        self.do_winds = False
        self.do_contours = False
        self.varconds = None
        self.grid = None

        if ('winds' in kwargs) \
                and (kwargs['winds']['src']) \
                and (kwargs['winds']['opts']):
            print "Will do winds"
            self.do_winds = True
            self.windsrc =  kwargs['winds']['src']
            self.windopts = kwargs['winds']['opts']

        if ('contours' in kwargs) \
                and (kwargs['contours']['var']):
            print "Will do contours"
            self.do_contours = True
            self.contoursrc  = self.srcs[0]
            self.contourvar  = kwargs['contours']['var']

        if ('varconds' in kwargs):
            vc = kwargs['varconds']
            if vc != None:
                self.varconds = eval(vc)

        #self.galab.cmd("reinit")
        # Open sources
        for f in self.srcs:
            #print "Opening data file", f
            if os.path.isabs(f):
                fpath = f
            else:
                fpath = os.path.join(self.indir, f)
            fin = nc(fpath)
            self.ncdf[0].append(fin)

        for v in fin.variables:
            if v == 'lon' or v == 'longitude':
                x = fin.variables[v][:]
            elif v == 'lat' or v == 'latitude':
                y = fin.variables[v][:]

        self.grid = (x, y)

        # Open wind source if available
        for w in self.windsrc:
            #print "Opening winds file", w
            if os.path.isabs(w):
                wpath = w
            else:
                wpath = os.path.join(self.indir, w)
            self.ncdf[1].append(nc(wpath))

#        if self.transf:
#            self.ncdf.transf = self.transf

    def ncdfReset(self):
        # close src files and winds
        for f in self.ncdf[0]:
            f.close()
        for w in self.ncdf[1]:
            w.close()
        print "Resetting GraDs engine"

    def getDataForTstep(self, step):
        self.data.reset()
        #print "Getting map data for step ", step
        self.data.setMapData(self.getMapDataForTstep(step))

        if self.do_winds:
            print "Getting wind data for step ", step
            self.data.setWindData(self.getWindDataForTstep(step))

        if self.do_contours:
            print "Getting contour data for step ", step
            self.data.setContourData(self.getContourDataForTstep(step))

        # Calculate maximum over data
        # FIXME should this step be optional?
#        if (self.data.hasMaxData()):
#            self.data.setMaxData(maximum(self.data.getMaxData(), self.data.getMapData()))
#        else:
#            self.data.setMaxData(self.data.getMapData())
#        self.lastTstep = step
        return self.data

    def getCurrentMaxData(self):
        '''Get current max data.
           NOTE: A call to this method will
           reset current max data, so it can be called only once! '''
        print "Retrieving max (last calculated tstep: %d); this will reset current MAX data!!" % self.lastTstep
        max_data = MapData(map_data=self.data.getMaxData())
        self.data.maxReset()
        return max_data

    def getMapDataForTstep(self, step):
        mapData = []
        for src_n in range(len(self.srcs)):
            tstep = step + int(self.srcgaps[src_n])
            srcfile = self.ncdf[0][src_n]
            #TODO zoom on lat lon according to the domain
#            self.ncdf.cmd("set dfile %d" % (src_n + 1))
#            self.ncdf.cmd("set lon %s %s" % (str(self.lons[0]), str(self.lons[1])))
#            self.ncdf.cmd("set lat %s %s" % (str(self.lats[0]), str(self.lats[1])))
#            self.ncdf.cmd("set t %s" % (TSTEP))
            # variable operations
#            if self.varconds and len(self.varconds)>=src_n:
#                varcond = self.varconds[src_n]
#                cond = varcond[0]
#                term = varcond[1]
#                vres = varcond[2]
#                if cond == '=':
#                    if TSTEP == int(term):
#                        EXPVAR = vres
#                elif cond == '<=':
#                    if TSTEP <= int(term):
#                        EXPVAR = vres
#                elif cond == '<':
#                    if TSTEP < int(term):
#                        EXPVAR = vres
#                elif cond == '>=':
#                    if TSTEP >= int(term):
#                        EXPVAR = vres
#                elif cond == '>':
#                    if TSTEP > int(term):
#                        EXPVAR = vres

            # variable with possible expression to calculate
            expvar = self.srcvars[src_n]
            try:
                # if no expression
                values = srcfile.variables[expvar][tstep,:,:]
            except Exception, e:
                # otherwise search for a variable of the dataset into the
                # expression given
                for v in srcfile.variables:
                    if v in expvar:
                        expvar = expvar.replace(v, 'newvar')
                        newvar = srcfile.variables[v][tstep,:,:]
                        values = eval(expvar)
                        break
            if mapData:
                #exp_res = self.ncdf.exp(EXPVAR)
                mapData.append(values)
            else:
                mapData = [values]

        return mapData

    def getWindDataForTstep(self, step):
        windData = {'u': None, 'v': None}
        self.ncdf.cmd("set dfile %d" % (len(self.srcs) + 1))
        self.ncdf.cmd("set lon %s %s" % (str(self.lons[0]), str(self.lons[1])))
        self.ncdf.cmd("set lat %s %s" % (str(self.lats[0]), str(self.lats[1])))
        self.ncdf.cmd("set t %s" % (step + 1))
        #self.ncdf.cmd("set lev 2932") #FIXME
        print "WINDOPTS", self.windopts
        windData['u'] = self.ncdf.exp("skip(%s,%s,%s)" % (self.windopts[0], self.windopts[2], self.windopts[2]))
        windData['v'] = self.ncdf.exp("%s" % self.windopts[1])
        if len(self.windopts) == 4:
            windData['barbs'] = self.windopts[3]
        return windData

    def getContourDataForTstep(self, step):
        cdata = None
        print "Contours variable: ", self.contourvar
        self.ncdf.cmd("set dfile %d" % 1)
        self.ncdf.cmd("set lon %s %s" % (str(self.lons[0]), str(self.lons[1])))
        self.ncdf.cmd("set lat %s %s" % (str(self.lats[0]), str(self.lats[1])))
        self.ncdf.cmd("set t %s" % (step + 1))
        cdata = self.ncdf.exp(self.contourvar)
        print "**************", self.lons, self.lats, step, self.contourvar
        print cdata
#        sys.exit(1)
        return cdata

    def getDims(self):
        print "Getting dimensions from file ", self.srcs[0]
        #self.ncdf.cmd("set dfile 1")
        time = self.ncdf[0][0].variables['time']
        typ, since, dat, tim = time.units.split()
        t = time[:]
        tim = tim.split('.')[0]
        tfmt = '%HZ%d%b%Y'
        try:
            dt = datetime.strptime("%s %s" % (dat,tim), '%Y-%m-%d %H:%M:%S')
        except:
            dt = datetime.strptime("%s %s" % (dat,tim), '%Y-%m-%d %H:%M')
        if typ == 'minutes':
            d = {(np.where(t==i)[0][0], (dt + timedelta(minutes=int(i))).strftime(tfmt)) for i in t}
        elif typ == 'seconds':
            d = {(np.where(t==i)[0][0], (dt + timedelta(seconds=int(i))).strftime(tfmt)) for i in t}
        else: #typ == 'hours':
            d = {(np.where(t==i)[0][0], (dt + timedelta(hours=int(i))).strftime(tfmt)) for i in t}
        return dict(d)


class CsvDataHandler:
    datafile = None
    separator = None
    fields = []

    def __init__(self, fname, separator=','):
        print 'opening ', fname
        self.datafile = fname
        self.separator = separator
        self.reader = None

    def filter(self, date, hour, **kwargs):
        lat = []
        lon = []
        val = []

        datecol = kwargs['datecol'] or 'date'
        hourcol = kwargs['hourcol'] or 'hour'
        latcol  = kwargs['latcol'] or 'lat'
        loncol  = kwargs['loncol'] or 'lon'
        valcol  = kwargs['valcol'] or 'val'

        self.reader = pd.read_csv(self.datafile, sep=self.separator)

        filtered = self.reader[(self.reader[datecol]==date.strip()) &
(self.reader[hourcol]==int(hour))]

        print 'filtering on date:', date, ' and hour', hour
        return filtered[latcol], filtered[loncol], filtered[valcol]


class DataTransform(object):
    @staticmethod
    def setColorsFromBuckets(colors, bounds, values):
        '''Returns a vector of the same size as values,
        containing a value picked from colors when the
        current value falls in a bucket as specified in
        bounds.

        Colors and bounds must have the same size'''

        print 'bounds:', bounds
        print 'colors:', colors
        print 'values:', values

        if (len(colors) != len(bounds) + 1):
            raise Exception("Wrong size for colors/bounds")

        out = []

        for idx in range(0, len(values)):
            print "processing", values[idx]
            if(values[idx] < bounds[0]):
                out.append(colors[0])
            elif(values[idx] >= bounds[len(bounds) - 1]):
                out.append(colors[len(colors) - 1])
            else:
                for bidx in range(0, len(bounds) - 1):
                    if(values[idx] >= bounds[bidx] and values[idx] < bounds[bidx + 1] ):
                        out.append(colors[bidx + 1])

        print out
        return out


class MapDrawOptions(object):
    """ """

    def __init__(self, **kwargs):
        ## General Options
        self.area_thresh = 500      # Default area threshold for maps
        self.keep_aspect = True     # Do not keep original proportion by default
        self.resolution = 'i'       # Default resolution is intermediate
        self.xsize = 1              # (when keep_aspect = False) Default horizontal size [value between 0 and 1]
        self.ysize = 1              # (when keep_aspect = False) Default vertical size [value between 0 and 1]
        self.dpi = 100              # Default DPI
        self.draw_colorbar = True   # Default policy is to draw colorbar
        self.coords_linewidth = 0.3
        self.coords_fontsize = 10
        self.coords_color = 'grey'

        ## FILLED COUNTOUR OPTIONS

        ## AIR QUALITY OPTIONS
        self.limits = []         # Default is no limits arrows

        ## WINDS OPTIONS
        self.wind_units = 'inches'    # Wind arrow default unit
        self.wind_scale = 40          # Wind arrow scale default
        self.wind_head_length = 10    # Wind arrowhead length default
        self.wind_head_width  = 8     # Wind arrowhead width default
        self.wind_width = 0.008       # Wind arrow width default
        self.wind_minshaft = 0.25     # Wind arrow minimum shaft default
        self.wind_label_xpos = 0.8    # Wind label x position (relative to quiver plot)
        self.wind_label_ypos = -0.075 # Wind label y position (relative to quiver plot)
        self.wind_label_scale = 20    # Size of the label (represents m/s)

        ## CONTOUR OPTIONS
        self.contours_int = 1      # Default contour interval
        self.contours_color = 'k' # Default contour color
        self.contours_linewidth = 0.8   # Default line width for contours
        self.contours_label = False # if True labels of contours are showed
        self.contours_label_fontsize = 13      # Default font size for contour label
        self.contours_label_format = '%d'     # Default format for contour label
        self.contours_exclude_vals = [0]    # Default list of values to exclude from drawing

        self.nocontourf = False # if True contourf is hidden
        self.nomap = False # if True map is hidden
        self.nointerp = False # if False use contourf, else pcolor

        ## SHAPEFILES
        self.shapefiles = []            # Default shapefiles list is empty
        self.shapef_width_init = 0.75   # Initial line width for shapefiles
        self.shapef_width_step = 0.25   # Line width difference between two consecutive shapefiles

    def hasShapeFiles(self):
        return (len(self.shapefiles) > 0)

    def hasLimits(self):
        return (len(self.limits) > 0)


class MapGenerator(object):
    """ Root class for plotting"""

    def __init__(self, **kwargs):
        """ Initialize MapGenerator with attributes common to all the plotting
        services """
        self.mgplot         = kwargs.get('mgplot', None)
        self.mgaxis         = kwargs.get('mgaxis', None)
        self.indir          = kwargs.get('indir', '.')
        self.outdir         = kwargs.get('outdir', '.')
        self.srcgaps        = kwargs.get('srcgaps', [0])
        self.title          = kwargs.get('title', '')
        self.fontsize       = kwargs.get('fontsize', 14)
        self.aspect         = kwargs.get('aspect', True)
        self.projection     = kwargs.get('projection', 'cyl')
        self.transf         = kwargs.get('transf', False)
        self.scatter        = kwargs.get('scatter', None)
        self.logo           = kwargs.get('logo', None)
        self.overwrite      = kwargs.get('overwrite', True)
        self.anim           = kwargs.get('anim', False)
        self.anim_delay     = kwargs.get('anim_delay', 25)
        self.img_template   = kwargs.get('img_template', '')
        self.joint_template = kwargs.get('joint_template', '')
        self.varconds       = kwargs.get('varconds', None)
        self.filefmt        = kwargs.get('filefmt', 'png')
        self.config_dir     = kwargs.get('config_dir', \
                                os.path.join(os.environ['HOME'], '.mapgenerator'))
        self.config_file    = kwargs.get('config_file', 'config.cfg')


class MapCross(MapGenerator):
    """ Define common attributes to maps and cross sections """

    def __init__(self, **kwargs):
        """ Initialize class with MapGenerator attributes plus some others """
        MapGenerator.__init__(self, **kwargs)
        self.bounds         = kwargs.get('bounds', [])
        self.colors         = kwargs.get('colors', 'jet')
        self.over           = kwargs.get('over', None)
        self.under          = kwargs.get('under', None)
        self.bad            = kwargs.get('bad', 'gray')
        self.lat            = kwargs.get('lat', ())
        self.lon            = kwargs.get('lon', ())
        self.wind           = kwargs.get('wind', [])
        self.windopts       = kwargs.get('windopts', [])
        self.contours       = kwargs.get('contours', None)
        self.contours_color = kwargs.get('contours_color', None)
        self.contours_label = kwargs.get('contours_label', None)
        self.contours_int   = kwargs.get('contours_int', 1)
        self.nointerp       = kwargs.get('nointerp', False)
        self.colorbar       = kwargs.get('colorbar', True)
        self.noruntime      = kwargs.get('noruntime', False)
        self.timesteps      = kwargs.get('timesteps', [0])
        self.limits         = kwargs.get('limits', None)

